import bpy
import bmesh
from .slt_utils import *


class SLT_OT_recalculate_loop(bpy.types.Operator):
    bl_idname = "smart_loops_toolkit.recalculate_loop"
    bl_label = "Recalculate Loop"
    bl_description = 'Recaululate all the verts in between the selected vertices so that they match the profile of a bezier curve'
    bl_options = {'REGISTER', 'UNDO'}


    closed_loop: bpy.props.BoolProperty(
        name='Closed Loop',
        description='Whether to also move the vertices between the first and last vert',
        default=False
    )

    project_mode: bpy.props.EnumProperty(
        name='Projection mode',
        description='Projection mode for use when projecting on face planes',
        items=[
            ('none', 'None', 'Do not to project the resulting vertices location on the plane of faces they modify. Often checked if used with singular holes'),
            ('nearby normal', 'Nearby normal', 'Project on faces that have closest normal to normal of the plane including the nearest verts'),
            ('selection normal', 'Selection normal', 'Project on faces that have closest normal to selection normal. This is not ideal for holes in extremely curved surfaces'),
            ('distance', 'Distance', 'Project on faces that are closer to the goal location. This is not ideal for surfaces with irregular bend (AR15 rail example)')
        ],
        default='nearby normal'
    )

    bezier_resolution: bpy.props.IntProperty(
        name='Bezier resolution',
        description='Determines the quality of the curve for approximating the position of vertices',
        default=20,
        min=1
    )

    # TODO: maybe in the future projection along edges as well? that's gonna be a lot of work

    def execute(self, context):
        obj = bpy.context.object

        # bisect edges
        bm = get_bmesh(obj, True, True, True)

        verts = [x for x in bm.select_history if isinstance(x, bmesh.types.BMVert)]
        curve_points = self.extract_curve_geo(obj, verts, self.bezier_resolution, self.closed_loop)

        if 'selection normal' in self.project_mode:
            try:
                path_node = dijkstra(bm, verts[0], verts[-1])[verts[-1]]
                shortest_path = list(set(flattened([e.verts for e in path_node.shortest_path])))
                global_selection_normal = get_selection_normal([v.co for v in shortest_path], [v.normal for v in shortest_path])
            except:
                show_message('Invalid selection.')
                free_bmesh(obj, bm)
                return {'FINISHED'}

        if self.closed_loop:
            verts.append(verts[0])

        moved_verts_indices = list()
        for i in range(1, len(verts)):
            path_node = dijkstra(bm, verts[i - 1], verts[i])[verts[i]]
            shortest_path = [e for e in path_node.shortest_path]

            if len(shortest_path) == 0:
                show_message('Invalid selection.')
                free_bmesh(obj, bm)
                return {'CANCELLED'}

            path_length = path_node.length

            corresponding_curve_verts = curve_points[(i - 1) * (self.bezier_resolution + 1) : i * (self.bezier_resolution + 1) + 1]

            current_vert = verts[i - 1]
            current_distance = 0
            new_positions = dict()
            this_path_verts_indices = [current_vert.index]
            for e in shortest_path[:len(shortest_path) - 1]:
                current_distance += e.calc_length()
                current_vert = other_edge_vertex(current_vert, e)

                # prevent moving a vert twice (in case of a closed loop sometimes the path between the last and first is the same as between first and last)
                if current_vert.index in moved_verts_indices or current_vert.select:
                    continue
                
                this_path_verts_indices.append(current_vert.index)

                normalized_distance = current_distance / path_length

                max_curve_index = len(corresponding_curve_verts) - 1
                curve_vert_index = int(normalized_distance * max_curve_index)

                granularity = 1.0 / max_curve_index
                remainder = (normalized_distance % granularity) / granularity

                new_position = corresponding_curve_verts[curve_vert_index] + remainder * (corresponding_curve_verts[curve_vert_index + 1] - corresponding_curve_verts[curve_vert_index])
                new_positions[current_vert] = new_position
            this_path_verts_indices.append(other_edge_vertex(current_vert, shortest_path[-1]).index)

            if len(shortest_path) < 2:
                continue

            possible_faces = get_relevant_faces(bm, this_path_verts_indices)
            local_normal = get_selection_normal(corresponding_curve_verts, [bm.verts[i].normal for i in this_path_verts_indices])
            for v, pos in new_positions.items():
                moved_verts_indices.append(v.index)
                if self.project_mode != 'none':
                    candidates = [f for f in possible_faces if v in f.verts]

                    # if no faces around: just use the position sampled from curve
                    if len(candidates) == 0:
                        v.co = pos
                        continue

                    if 'nearby normal' in self.project_mode:
                        face = min(candidates, key=lambda f: (local_normal - f.normal).length)
                    elif 'selection normal' in self.project_mode:
                        face = min(candidates, key=lambda f: (global_selection_normal - f.normal).length)
                    elif 'distance' in self.project_mode:
                        face = min(candidates, key=lambda f: point_to_plane_dist(pos, f.calc_center_median(), f.normal))
                    
                    v.co = project_point_on_plane(pos, face.calc_center_median(), face.normal)
                else:
                    v.co = pos

        update_bmesh(obj, bm)
        free_bmesh(obj, bm)

        return {'FINISHED'}


    @staticmethod
    def extract_curve_geo(obj, verts:list[bmesh.types.BMVert], cuts:int, cyclic:bool):
        curve_vertpos = [v.co for v in verts]
        # make a new curve
        crv = bpy.data.curves.new('crv', 'CURVE')
        crv.dimensions = '3D'
        crv.resolution_u = cuts + 1

        # BEZIER:
        # make a new spline in that curve
        spline = crv.splines.new(type='BEZIER')

        # a spline point for each point
        spline.bezier_points.add(len(curve_vertpos) - 1) # theres already one point by default

        # assign the point coordinates to the spline points
        for i in range(len(curve_vertpos)):
            spline.bezier_points[i].co = curve_vertpos[i]
            spline.bezier_points[i].handle_left = spline.bezier_points[i].handle_right = spline.bezier_points[i].co
            spline.bezier_points[i].handle_left_type = 'AUTO'
            spline.bezier_points[i].handle_right_type = 'AUTO'
        
        spline.use_cyclic_u = cyclic

        # make a new object with the curve
        obj_curve = bpy.data.objects.new('helper_curve', crv)
        mesh_curve = obj_curve.to_mesh()

        return [v.co.copy() for v in mesh_curve.vertices]
