import bpy
from .slt_utils import *


class SLT_OT_straighten_loops(bpy.types.Operator):
    bl_idname = "smart_loops_toolkit.straighten_loops"
    bl_label = "Straighten Loops"
    bl_description = 'Straighten loops by projecting all the vertices on the line created by the tips of the loop'
    bl_options = {'REGISTER', 'UNDO'}

    straighten_mode: bpy.props.EnumProperty(
        name='Straighten mode',
        description='Straighten mode for use when straightening verts',
        items=[
            ('simple', 'Simple', 'Just project the vertices on the edge'),
            ('edges', 'Along edges', 'Move the vertices along edges to straighten verts')
        ],
        default='edges'
    )

    dissolve_inner: bpy.props.BoolProperty(
        name='Dissolve inner verts',
        description='Whether or not to make a singular edge out of the straightened loop',
        default=False
    )

    def execute(self, context):
        obj = bpy.context.object

        # bisect edges
        bm = get_bmesh(obj, True, True, True)
        selected_edges = [e for e in bm.edges if e.select]

        if len(selected_edges) < 1:
            show_message('Invalid selection. Please select more than one edge.')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        try:
            loops = get_separated_sorted_loops(selected_edges)
        except:
            show_message('Invalid selection.')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        for loop in loops:
            first = bm.verts[loop[0]].co
            last = bm.verts[loop[-1]].co

            for p in loop[1:len(loop) - 1]:
                vert = bm.verts[p]

                if self.straighten_mode == 'simple':
                    vert.co = project_point_on_line(vert.co, first, last)
                elif self.straighten_mode == 'edges':
                    possible_line_verts = [other_edge_vertex(vert, e) for e in vert.link_edges if e.verts[0].index not in loop or e.verts[1].index not in loop]

                    if len(possible_line_verts) > 0:
                        # 'average out' the normal by crossing with all the outgoing verts
                        normal = possible_line_verts[0].co - vert.co
                        for v in possible_line_verts[1:]:
                            normal = normal.cross(v.co - vert.co).normalized()
                        # if we cross uneven number of times, the normal is going to be wrong, so also cross with the vector perpendicular to the loop direction
                        if len(possible_line_verts) % 2 == 1:
                            perp_line = ((last - first).cross(vert.co - first)).normalized()
                            normal = normal.cross(perp_line).normalized()
                        intersection = line_plane_intersection(first, last, vert.co, normal)
                        if intersection is None or not is_projected_point_on_segment(intersection, first, last):
                            vert.co = project_point_on_line(vert.co, first, last)
                        else:
                            vert.co = intersection
                    else:
                        vert.co = project_point_on_line(vert.co, first, last)

            if self.dissolve_inner:
                v1, v2 = bm.verts[loop[0]], bm.verts[loop[-1]]
                bmesh.ops.dissolve_verts(bm, verts=[bm.verts[i] for i in loop[1:len(loop) - 1]])
                bm.verts.ensure_lookup_table()
                bm.edges.ensure_lookup_table()
                bm.faces.ensure_lookup_table()
                for e in bmesh.ops.connect_verts(bm, verts=[v1, v2])['edges']:
                    e.select = True

        update_bmesh(obj, bm)
        free_bmesh(obj, bm)

        return {'FINISHED'}
