import bpy
from itertools import combinations

from mathutils import Vector

from .slt_utils import *
from bmesh.types import BMVert


class SLT_OT_unbevel(PrecisionOperatorMixin, bpy.types.Operator):
    bl_idname = "smart_loops_toolkit.unbevel"
    bl_label = "Unbevel"
    bl_description = 'Unbevel loops'
    bl_options = {'REGISTER', 'UNDO'}

    def execute_with_precision(self, context: bpy.context) -> typing.Union[typing.Set[int], typing.Set[str]]:
        obj = bpy.context.object

        bm = get_bmesh(obj, True, True, True)
        selected_edges = [e for e in bm.edges if e.select]

        # invalid selection handling
        if len(selected_edges) < 2:
            show_message('Invalid selection.')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        # selected_verts = [v for v in bm.verts if v.select]
        # grouped_loops = group_loops(loops, get_relevant_faces(bm, [v.index for v in selected_verts])) # TODO: maybe support multiple loop unbeveling at once <- the solution on the left does not keep the correct order...
        try:
            all_loops = get_separated_sorted_loops_bevel(bm, selected_edges)
        except:
            show_message('Invalid selection. Select at least two loops linked by edges (do not select the linking edges)!')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        if len(all_loops) < 2:
            show_message('Invalid selection. Select at least two loops linked by edges (do not select the linking edges)!')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        if any([len(loop) != len(all_loops[0]) for loop in all_loops]):
            show_message('Invalid selection. Selected loops of different length!')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        # assign boundary loops
        boundary_loops = [all_loops[0], all_loops[-1]]
        loop_next_to_boundary = [all_loops[1], all_loops[-2]]

        # evaluate the target locations for each loop
        vert_index_outer_planes = dict()
        vert_index_inner_planes = dict()
        for i in range(len(boundary_loops[0])):
            for j in range(2):
                closest_faces = sorted(bm.verts[boundary_loops[j][i]].link_faces,
                                       key=lambda f: ([v.index in boundary_loops[j] for v in f.verts].count(True),
                                                      -[v.index in loop_next_to_boundary[j] for v in f.verts].count(True)), reverse=True)

                outer_options = [f for f in closest_faces if not any([v.index in loop_next_to_boundary[j] for v in f.verts])]
                vert_index_outer_planes[boundary_loops[j][i]] = [Plane(center=f.verts[0].co, normal=f.normal, index=f.index) for f in outer_options]
                vert_index_inner_planes[boundary_loops[j][i]] = [Plane(center=f.verts[0].co, normal=f.normal, index=f.index) for f in closest_faces if f not in outer_options]

        target_locations = list()
        keys = list(vert_index_outer_planes.keys())
        for i in range(0, len(keys), 2):
            vert_a_index = keys[i]
            vert_b_index = keys[i + 1]

            planes_tested = list()
            for x in vert_index_outer_planes[vert_a_index] + vert_index_outer_planes[vert_b_index]:
                if x not in planes_tested: # kinda non-pythonic method but Plane can't be reasonably hashable
                    planes_tested.append(x)

            # 0 planes tested - need to use inner planes
            if (len(planes_tested) == 0):
                for x in vert_index_inner_planes[vert_a_index] + vert_index_inner_planes[vert_b_index]:
                    if x not in planes_tested: # kinda non-pythonic method but Plane can't be reasonably hashable
                        planes_tested.append(x)

            # bevel dir for calculating cross products
            bevel_dir = bm.verts[vert_b_index].co - bm.verts[vert_a_index].co
            are_planes_parallel = False

            # 1 plane available: parallel planes - figure out the remaining planes via cross products
            if len(planes_tested) == 1:
                print("All planes parallel (single plane)")
                are_planes_parallel = True
                second_plane_normal = planes_tested[0].normal.cross(bevel_dir)
                planes_tested.append(Plane(center=bm.verts[vert_a_index].co, normal=second_plane_normal))
                planes_tested.append(Plane(center=bm.verts[vert_a_index].co, normal=planes_tested[0].normal.cross(second_plane_normal)))
            # 2 planes available: figure out the 3rd plane via cross product
            elif len(planes_tested) == 2:
                if vectors_equal_abs(planes_tested[0].normal, planes_tested[1].normal):
                    print("Two planes but parallel")
                    are_planes_parallel = True
                    # TODO: do the same as with bevel dir above but for bevel dir should use the vert linked with vertA (one that would merge with)
                    planes_tested.pop(1)
                    second_plane_normal = mathutils.Vector((0.238472384, 0.374831, -0.348243)).normalized() # <- kinda messy workaround before implementing the above TODO
                    planes_tested.append(Plane(center=bm.verts[vert_a_index].co, normal=second_plane_normal))
                    planes_tested.append(Plane(center=bm.verts[vert_a_index].co, normal=planes_tested[0].normal.cross(second_plane_normal)))
                else:
                    print("Just two planes")
                    planes_tested.append(Plane(center=bm.verts[vert_a_index].co, normal=planes_tested[0].normal.cross(bevel_dir)))
            else:
                print("More than two planes")

            def calculate_differing_factor(planes: list[Plane]) -> float:
                return planes[0].normal.angle(planes[1].normal) + planes[1].normal.angle(planes[2].normal) + planes[0].normal.angle(planes[2].normal) # planes[0].normal.cross(planes[1].normal).magnitude * planes[1].normal.cross(planes[2].normal).magnitude * planes[0].normal.cross(planes[2].normal).magnitude

            # get most-differing set of planes in terms of direction
            best_fit_planes = max(combinations(planes_tested, 3), key=calculate_differing_factor)

            # print(list(zip([[y.index for y in x] for x in plane_options], list(map(most_differing_vectors, plane_options)))))

            target_locations.append((are_planes_parallel, three_plane_intersection(best_fit_planes[0], best_fit_planes[1], best_fit_planes[2])))

        # TODO: this tool can be enhanced quite a bit still:
        #  1. we could move the loops to the "average" bevel vertex position (or center vertex? or furthest point?) if one plane parallel
        #  2. beveling a thing with no outer faces gives a single edge
        #  3. two separate and parallel planes (cube with top double-beveled) just cuts the top - we could offset it to the furthest point
        #  4. probs there are a few cases in which the tool bugs out for no good reason

        # create vert merging list
        merged_verts: list[list[BMVert]] = list()
        for i in range(len(target_locations)):
            merged_verts.append([bm.verts[loop[i]] for loop in all_loops])

        # merge verts accordingly
        for i in range(len(target_locations)):
            if target_locations[i][0]:
                # if parallel
                bmesh.ops.pointmerge(bm, verts=merged_verts[i][:-1], merge_co=target_locations[i][1])
            else:
                bmesh.ops.pointmerge(bm, verts=merged_verts[i], merge_co=target_locations[i][1])

        update_bmesh(obj, bm)
        free_bmesh(obj, bm)

        return {'FINISHED'}
