import bpy
import bmesh
from bmesh.types import BMVert
from mathutils import Vector

from .slt_utils import get_bmesh, get_separated_sorted_loops_bevel, show_message, free_bmesh, edge_exists, update_bmesh


class SLT_OT_unsubdivide_loops(bpy.types.Operator):
    bl_idname = "smart_loops_toolkit.unsubdivide_loops"
    bl_label = "Unsubdivide Loops"
    bl_description = 'Unsubdivide loops to create a low detailed version of the selection'
    bl_options = {'REGISTER', 'UNDO'}

    level: bpy.props.IntProperty(name='Level',
                                 description='How many verts to merge into one',
                                 default=2, min=2)

    offset: bpy.props.IntProperty(name='Offset',
                                  description='Initial offset to start merging vertices',
                                  default=0)

    def execute(self, context):
        obj = bpy.context.object

        bm = get_bmesh(obj, True, True, True)
        selected_edges = [e for e in bm.edges if e.select]

        try:
            loops = get_separated_sorted_loops_bevel(bm, selected_edges, allow_single_loop=True)
        except Exception as e:
            self.report({'ERROR'}, f'Invalid selection. Remember to only select loops, not full faces. Error: {e}')
            free_bmesh(obj, bm)
            return {'CANCELLED'}

        to_pointmerge: list[tuple[list[BMVert], Vector]] = list()
        max_verts_count = min((len(l) for l in loops))
        offset_normalized = self.offset % max_verts_count
        for loop in loops:
            if edge_exists(bm.verts[loop[0]], bm.verts[loop[-1]]):  # closed loop
                print('cyclic')
                for i in range(offset_normalized, offset_normalized + max_verts_count - 1, self.level):
                    start = i % max_verts_count
                    print([j % max_verts_count for j in range(start, start + self.level)])
                    to_merge = [bm.verts[loop[j % max_verts_count]] for j in range(start, start + self.level)]
                    merge_pos = bm.verts[loop[start]].co
                    to_pointmerge.append((to_merge, merge_pos))
            else:  # open loop
                # handle part from offset towards the end
                print('forwards')
                for i in range(offset_normalized, max_verts_count - 1, self.level):
                    print([j for j in range(i, min(i + self.level, max_verts_count - 1))])
                    to_merge = [bm.verts[loop[j]] for j in range(i, min(i + self.level, max_verts_count - 1))]
                    if len(to_merge) > 1:
                        merge_pos = bm.verts[loop[i]].co
                        to_pointmerge.append((to_merge, merge_pos))
                # handle part from the offset towards the start
                print('backwards')
                for i in range(offset_normalized, 1, -self.level):
                    print([j for j in range(i, max(i - self.level, 0), -1)])
                    to_merge = [bm.verts[loop[j]] for j in range(i, max(i - self.level, 0), -1)]
                    if len(to_merge) > 1:
                        merge_pos = bm.verts[loop[i]].co
                        to_pointmerge.append((to_merge, merge_pos))

        try:
            for x in to_pointmerge:
                bmesh.ops.pointmerge(bm, verts=x[0], merge_co=x[1])
        except Exception:
            update_bmesh(obj, bm)
            free_bmesh(obj, bm)
            self.report({"INFO"}, "Encountered an error when unsubdividing: is the level value too big?")
            return {'CANCELLED'}

        update_bmesh(obj, bm)
        free_bmesh(obj, bm)

        return {'FINISHED'}
